from __future__ import annotations

import importlib
import importlib.util
import inspect
import os
import re
import sys
from enum import Enum
from typing import Any

from ._component import Component, Provider
from ._models import Operation
from .constants import ROOT_PACKAGE_NAME
from .dependency import ComponentNode, ProviderNode
from .exceptions import LoadError
from .manifest import ComponentInstance, Manifest
from .spec import Kind


class Loader:
    MANIFEST_FILE = "manifest.yaml"
    path: str
    manifest_path: str

    manifest: Manifest
    root: str

    _init = False
    _dependencies: dict[str, ComponentDependency] = dict()

    def __init__(
        self,
        path: str = ".",
        manifest: str = MANIFEST_FILE,
        handle: str | None = None,
    ):
        self.path = path
        self.manifest_path = os.path.join(
            path,
            manifest,
        )
        self.manifest = Manifest.parse(path=self.manifest_path)
        if handle is not None:
            self.root = handle
        elif self.manifest.root is not None:
            self.root = self.manifest.root
        abs_project_folder = os.path.abspath(path)
        if abs_project_folder not in sys.path:
            sys.path.append(abs_project_folder)
        self._prepare_unresolved()

    def generate_requirements(
        self,
        handle: str | None = None,
        out: str | None = None,
    ) -> dict[str, list[str]]:
        requirements: dict[str, list[str]] = {"verse-core": [], "verse-ql": []}

        def traverse(node: ComponentNode):
            cnamespace, cpackage = node.name.split(".", 1)
            if node.kind == Kind.CUSTOM or node.provider.kind == Kind.CUSTOM:
                pass
            else:
                package = f"verse-{cnamespace}"
                extra = (
                    cpackage.replace("_", "-")
                    + "-"
                    + node.provider.name.replace("_", "-")
                )
                if package not in requirements:
                    requirements[package] = []
                if extra not in requirements[package]:
                    requirements[package].append(extra)

            for child in node.depends:
                traverse(child)
            for child in node.provider.depends:
                traverse(child)

        graph = self.get_dependency_graph(handle)
        traverse(graph)

        if out is not None:
            content = ""
            for package in requirements:
                if len(requirements[package]) > 0:
                    content += (
                        package
                        + "["
                        + str.join(",", requirements[package])
                        + "]"
                        "\n"
                    )
                else:
                    content += package + "\n"
            with open(out, "w") as file:
                file.write(content)
        return requirements

    def load_root(self) -> Component:
        if self.root is None:
            raise LoadError("Root handle must be specified")
        return self._resolve_component(self.root).component

    def load_component(self, handle: str) -> Component:
        return self._resolve_component(handle).component

    def get_dependency_graph(self, handle: str | None = None) -> ComponentNode:
        def _get_node(component: ComponentDependency) -> ComponentNode:
            node = ComponentNode(
                handle=component.handle,
                name=component.name,
                kind=(
                    Kind.CUSTOM
                    if component.name.startswith("custom.")
                    else Kind.BASE
                ),
                provider=ProviderNode(
                    name=component.provider.name,
                    kind=(
                        Kind.CUSTOM
                        if component.provider.name.startswith("custom.")
                        else Kind.BASE
                    ),
                ),
            )
            dependencies = self._get_dependencies_from_param(
                component.parameters
            )
            for dependency in dependencies:
                node.depends.append(_get_node(dependency))

            dependencies = self._get_dependencies_from_param(
                component.provider.parameters
            )
            for dependency in dependencies:
                node.provider.depends.append(_get_node(dependency))
            return node

        return _get_node(
            self._dependencies[self.root if handle is None else handle]
        )

    def _resolve_component(
        self,
        handle: str,
    ) -> ComponentDependency:
        if handle in self._dependencies:
            if self._dependencies[handle].resolved:
                return self._dependencies[handle]
            cdep = self._dependencies[handle]
        else:
            raise LoadError(f"Unresolved component {handle}")
        cdep.parameters = self._resolve_param(
            cdep.parameters,
        )
        pdep = self._resolve_provider(handle)
        cdep.component = self._init_component(cdep, pdep.provider)
        cdep.component._handle = handle
        cdep.resolved = True
        return cdep

    def _resolve_provider(
        self,
        component_handle: str,
    ) -> ProviderDependency:
        cdep = self._dependencies[component_handle]
        pdep = cdep.provider
        pdep.parameters = self._resolve_param(
            pdep.parameters,
        )
        pdep.provider = self._init_provider(pdep)
        pdep.resolved = True
        return pdep

    def _resolve_param(self, value: Any) -> Any:
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = self._resolve_param(v)
            return value
        elif isinstance(value, list):
            for i, item in enumerate(value):
                value[i] = self._resolve_param(item)
            return value
        elif isinstance(value, str) and self._is_ref(value):
            ref_info = self._get_ref_info(value)
            if ref_info.type == RefType.COMPONENT:
                return self._resolve_component(
                    ref_info.component_handle,
                ).component
            elif ref_info.type == RefType.COMPONENT_GET:
                return self._resolve_component_get(
                    self._resolve_component(
                        ref_info.component_handle,
                    ).component,
                    ref_info.get_args,
                )
        return value

    def _get_dependencies_from_param(
        self, value: Any
    ) -> list[ComponentDependency]:
        dependencies: list = []
        if isinstance(value, dict):
            for k, v in value.items():
                dependencies.extend(self._get_dependencies_from_param(v))
        elif isinstance(value, list):
            for i, item in enumerate(value):
                dependencies.extend(self._get_dependencies_from_param(item))
        elif isinstance(value, str) and self._is_ref(value):
            ref_info = self._get_ref_info(value)
            if ref_info.component_handle in self._dependencies:
                dependencies.append(
                    self._dependencies[ref_info.component_handle]
                )
            else:
                raise LoadError(
                    f"Unresolved component {ref_info.component_handle}"
                )
        return dependencies

    def _init_component(
        self,
        cdep: ComponentDependency,
        provider: Provider,
    ) -> Component:
        path = cdep.py_path
        if cdep.py_path is not None and cdep.py_path.startswith("."):
            path = os.path.join(
                self.path,
                cdep.name,
                cdep.py_path,
            )
        component = Loader.load_component_instance(
            path=path,
            parameters=cdep.parameters,
            provider=provider,
        )
        return component

    def _init_provider(self, pdep: ProviderDependency) -> Provider:
        path = pdep.py_path
        if pdep.py_path is not None and pdep.py_path.startswith("."):
            path = os.path.join(
                self.path,
                pdep.component_name,
                pdep.py_path,
            )
        provider = Loader.load_provider_instance(
            path=path,
            parameters=pdep.parameters,
        )
        return provider

    def _prepare_unresolved(self):
        for component in self.manifest.components:
            self._prepare_component_dependency(component)

    def _prepare_component_dependency(self, component: ComponentInstance):
        cdep = ComponentDependency()
        cdep.handle = component.handle
        cdep.name = component.name
        cdep.parameters = component.parameters
        cnamespace, cpackage = component.name.split(".", 1)
        if cnamespace == "custom":
            cdep.py_path = f"{cpackage}.component"
        else:
            cdep.py_path = (
                f"{ROOT_PACKAGE_NAME}.{cnamespace}.{cpackage}.component"
            )
        pdep = ProviderDependency()
        pdep.component_name = component.name
        pdep.name = component.provider.name
        pdep.parameters = component.provider.parameters
        if cnamespace == "custom" or component.provider.name.startswith(
            "custom."
        ):
            ppackage = component.provider.name.split(".")[-1]
            pdep.py_path = f"{cpackage}.providers.{ppackage}"
        else:
            cnamespace = f"{ROOT_PACKAGE_NAME}.{cnamespace}.{cpackage}"
            pdep.py_path = f"{cnamespace}.providers.{component.provider.name}"
        pdep.resolved = False
        cdep.provider = pdep
        cdep.resolved = False
        self._dependencies[component.handle] = cdep

    def _resolve_component_get(self, component: Component, get_args: dict):
        res = component.run(operation=Operation(name="get", args=get_args))
        return res.result.value

    def _resolve_provider_get(self, provider: Provider, get_args: dict):
        res = provider.run(operation=Operation(name="get", args=get_args))
        return res.result.value

    def _get_ref_info(self, str: str):
        pattern = r"^\{\{(.+)\}\}$"
        match = re.match(pattern, str)
        if not match:
            raise LoadError(f"{str} not a ref")
        value = match.group(1)
        splits = value.split(".")
        ref_info = RefInfo()
        ref_info.component_handle = splits[0]
        if len(splits) == 1:
            ref_info.type = RefType.COMPONENT
        elif len(splits) == 2:
            ref_info.type = RefType.COMPONENT_GET
            ref_info.get_args = dict(key=dict(id=splits[1]))
        elif len(splits) == 3:
            ref_info.type = RefType.COMPONENT_GET
            ref_info.get_args = dict(key=dict(label=splits[1], id=splits[2]))
        return ref_info

    def _is_ref(self, str: str) -> bool:
        pattern = r"^\{\{.+\}\}$"
        return bool(re.match(pattern, str))

    def _get_path(self, base_dir: str, manifest_path: str):
        return os.path.join(base_dir, manifest_path)

    @staticmethod
    def load_provider_instance(
        path: str | None = None,
        parameters: dict[str, Any] = dict(),
    ) -> Provider:
        if path is None:
            return Provider(**parameters)
        provider = Loader.load_class(path, Provider)
        return provider(**parameters)

    @staticmethod
    def load_component_instance(
        path: str | None,
        parameters: dict,
        provider: Provider,
    ) -> Component:
        if path is None:
            return Component(provider=provider, **parameters)
        component = Loader.load_class(path, Component)
        return component(provider=provider, **parameters)

    @staticmethod
    def load_class(path: str, type: Any) -> Any:
        class_name = None
        if ":" in path:
            splits = path.split(":")
            module_name = splits[0]
            class_name = splits[-1]
        else:
            module_name = path
        if module_name.endswith(".py"):
            import sys

            normalized_module_name = (
                os.path.normpath(module_name)
                .replace("\\", "/")
                .split(".py")[0]
                .replace("/", ".")
                .lstrip(".")
            )
            spec = importlib.util.spec_from_file_location(
                normalized_module_name, os.path.abspath(path)
            )
            if spec is not None:
                module = importlib.util.module_from_spec(spec)
                sys.modules[normalized_module_name] = module
                if spec.loader is not None:
                    spec.loader.exec_module(module)
            module_name = normalized_module_name
        else:
            module = importlib.import_module(module_name)
        if class_name is not None:
            return getattr(module, class_name)
        else:
            for name, cls in inspect.getmembers(module, inspect.isclass):
                if issubclass(cls, type) and cls.__module__ == module_name:
                    return cls
        raise LoadError(f"{type.__name__} not found at {module}")


class RefInfo:
    type: RefType
    component_handle: str
    get_args: dict


class RefType(int, Enum):
    COMPONENT = 0
    COMPONENT_GET = 1


class ComponentDependency:
    handle: str
    name: str
    parameters: dict[str, Any]
    py_path: str | None
    provider: ProviderDependency
    component: Component
    resolved: bool = False


class ProviderDependency:
    component_name: str
    name: str
    parameters: dict[str, Any]
    py_path: str | None
    provider: Provider
    resolved: bool = False



import argparse
from typing import Any

from ._loader import Loader
from ._models import Operation


def run(
    path: str,
    manifest: str,
    handle: str | None,
) -> Any:
    """
    Verse Run
    """
    loader = Loader(path=path, manifest=manifest, handle=handle)
    root_component = loader.load_root()
    operation = Operation(
        name="run",
        args=dict(path=path, manifest=manifest, root=handle),
    )
    return root_component.run(operation=operation)


def requirements(
    path: str,
    manifest: str,
    root: str | None,
    out: str | None,
):
    """
    Verse Requirements
    """
    loader = Loader(path=path, manifest=manifest, handle=root)
    requirements = loader.generate_requirements(out=out)
    return requirements


def main():
    parser = argparse.ArgumentParser(prog="verse", description="Verse CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)
    run_parser = subparsers.add_parser("run", help="Run the Verse application")
    requirements_parser = subparsers.add_parser(
        "requirements", help="Generate the pip requirements"
    )
    run_parser.add_argument(
        "--path",
        type=str,
        default=".",
        help="Project directory",
    )
    run_parser.add_argument(
        "--manifest",
        type=str,
        default=Loader.MANIFEST_FILE,
        help="Manifest filename",
    )
    run_parser.add_argument(
        "--handle",
        type=str,
        default=None,
        help="Root handle",
    )

    requirements_parser.add_argument(
        "--path",
        type=str,
        default=".",
        help="Project directory",
    )
    requirements_parser.add_argument(
        "--manifest",
        type=str,
        default=Loader.MANIFEST_FILE,
        help="Manifest filename",
    )
    requirements_parser.add_argument(
        "--handle",
        type=str,
        default=None,
        help="Root handle",
    )
    requirements_parser.add_argument(
        "--out",
        type=str,
        default=None,
        help="Output path",
    )

    args = parser.parse_args()
    if args.command == "run":
        run(
            path=args.path,
            manifest=args.manifest,
            handle=args.handle,
        )
    elif args.command == "requirements":
        requirements(
            path=args.path,
            manifest=args.manifest,
            root=args.handle,
            out=args.out,
        )
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import inspect
from functools import wraps
from typing import Any, Callable, TypeVar, cast, get_origin, get_type_hints

from ._models import Operation, Response

T = TypeVar("T", bound=Callable[..., Any])


def operation() -> Callable[[T], T]:
    def decorator(func: T) -> T:
        if not inspect.iscoroutinefunction(func):

            @wraps(func)
            def wrapper(*args, **kwargs) -> Any:
                self = args[0]
                context = kwargs.pop("context", None)

                sig = inspect.signature(func)
                bound_args = sig.bind(*args, **kwargs)
                bound_args.apply_defaults()
                locals = dict(bound_args.arguments)
                locals.pop("self", None)

                operation = Operation.normalize(
                    name=func.__name__,
                    args=locals,
                )
                response = self.run(operation, context)
                return_type = get_type_hints(func).get("return", None)
                if return_type is None:
                    return response
                elif get_origin(return_type) is Response:
                    return response
                else:
                    return response.result

            return cast(T, wrapper)
        else:

            @wraps(func)
            async def wrapper(*args, **kwargs) -> Any:
                self = args[0]
                context = kwargs.pop("context", None)

                sig = inspect.signature(func)
                bound_args = sig.bind(*args, **kwargs)
                bound_args.apply_defaults()
                locals = dict(bound_args.arguments)
                locals.pop("self", None)

                operation = Operation.normalize(
                    name=func.__name__[1:],
                    args=locals,
                )
                response = await self.arun(operation, context)
                return_type = get_type_hints(func).get("return", None)
                if return_type is None:
                    return response
                elif get_origin(return_type) is Response:
                    return response
                else:
                    return response.result

            return cast(T, wrapper)

    return decorator



from __future__ import annotations

from typing import Any

from ._models import DataModel
from ._yaml_loader import YamlLoader

__all__ = [
    "ComponentInstance",
    "Manifest",
    "ProviderInstance",
]


class ProviderInstance(DataModel):
    name: str
    parameters: dict[str, Any] = dict()


class ComponentInstance(DataModel):
    handle: str
    name: str
    parameters: dict[str, Any] = dict()
    provider: ProviderInstance


class Manifest(DataModel):
    components: list[ComponentInstance] = []
    root: str | None = None

    @staticmethod
    def parse(path: str) -> Manifest:
        obj = YamlLoader.load(path=path)
        manifest = Manifest.from_dict(obj)
        return manifest



from __future__ import annotations

from ._models import DataModel, Kind


class ComponentNode(DataModel):
    handle: str
    name: str
    kind: Kind
    provider: ProviderNode
    depends: list[ComponentNode] = []


class ProviderNode(DataModel):
    name: str
    kind: str
    depends: list[ComponentNode] = []



from typing import Any

from ._async_helper import AsyncHelper
from ._models import Context, Operation, Response
from .exceptions import NotSupportedError


class Provider:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def init(self, context: Context | None = None) -> None:
        pass

    async def ainit(self, context: Context | None = None) -> None:
        await AsyncHelper.to_async(func=self.init, context=context)

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Response[Any]:
        if operation is not None and operation.name is not None:
            func_name = operation.name
            func = getattr(self, func_name, None)
            if func is not None and callable(func):
                self.init()
                result = func(**operation.args or {})
                if not isinstance(result, Response):
                    return Response(result=result)
                return result

        raise NotSupportedError(
            operation.to_json() if operation is not None else None
        )

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Response[Any]:
        if operation is not None and operation.name is not None:
            func_name = f"a{operation.name}"
            func = getattr(self, func_name, None)
            if func is not None and callable(func):
                await self.ainit()
                result = await func(**operation.args or {})
                if not isinstance(result, Response):
                    return Response(result=result)
                return result

        return await AsyncHelper.to_async(
            func=self.run, operation=operation, context=context
        )



from __future__ import annotations

from enum import Enum

from ._models import DataModel, Kind
from ._yaml_loader import YamlLoader
from .constants import ROOT_PACKAGE_NAME
from .exceptions import NotFoundError

__all__ = [
    "Kind",
    "ComponentSpec",
    "ContentModal",
    "OperationArgSpec",
    "OperationExceptionSpec",
    "OperationReturnSpec",
    "OperationSpec",
    "ParameterOptionSpec",
    "ParameterSpec",
    "ProviderMode",
    "ProviderSpec",
    "PySpec",
    "TypePropertySpec",
    "TypeSpec",
]


class ContentModal(str, Enum):
    TEXT = "text"
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    MULTI = "multi"


class ProviderMode(str, Enum):
    MIXED = "mixed"
    LOCAL = "local"
    SERVICE = "service"


class ParameterOptionSpec(DataModel):
    label: str
    value: str | int | float | bool | None


class ParameterSpec(DataModel):
    name: str
    type: str
    description: str | None = None
    required: bool = True
    default: str | int | float | bool | dict | list | None = None
    options: list[ParameterOptionSpec] = []


class OperationArgSpec(DataModel):
    name: str
    type: str
    description: str | None = None
    required: bool = True
    default: str | int | float | bool | dict | list | None = None


class OperationReturnSpec(DataModel):
    type: str
    description: str | None = None


class OperationExceptionSpec(DataModel):
    type: str
    description: str | None = None


class OperationSpec(DataModel):
    name: str
    args: list[OperationArgSpec] = []
    description: str | None = None
    ret: OperationReturnSpec | None = None
    exceptions: list[OperationExceptionSpec] = []


class TypePropertySpec(DataModel):
    name: str
    type: str
    description: str | None = None
    required: bool | None = True
    default: str | int | float | bool | dict | list | None = None


class TypeSpec(DataModel):
    name: str
    description: str | None = None
    properties: list[TypePropertySpec] = []


class PySpec(DataModel):
    package: str | None = None
    path: str | None = None
    requirements: list[str] = []


class ProviderSpec(DataModel):
    name: str
    version: str | None = None
    title: str | None = None
    description: str | None = None
    logo: str | None = None
    author: str | None = None
    readme: str | None = None
    link: str | None = None
    kind: Kind = Kind.BASE
    mode: ProviderMode = ProviderMode.MIXED
    parameters: list[ParameterSpec] = []
    py: PySpec | None = None


class ComponentSpec(DataModel):
    name: str
    version: str | None = None
    title: str | None = None
    description: str | None = None
    logo: str | None = None
    author: str | None = None
    readme: str | None = None
    link: str | None = None
    category: str | None = None
    modals: list[ContentModal] = []
    kind: Kind = Kind.BASE
    parameters: list[ParameterSpec] = []
    operations: list[OperationSpec] = []
    providers: list[ProviderSpec] = []
    types: list[TypeSpec] = []
    py: PySpec | None = None

    @staticmethod
    def parse(path: str) -> ComponentSpec:
        obj = YamlLoader.load(path=path)
        manifest = ComponentSpec.from_dict(obj)
        return manifest

    def get_provider(self, name: str) -> ProviderSpec:
        for provider in self.providers:
            if name == provider.name:
                return provider
        raise NotFoundError("Provider not found")

    def get_py_path(self) -> str | None:
        if self.py is not None and self.py.path is not None:
            return self.py.path
        if self.kind == Kind.CUSTOM:
            return "./component.py"
        namespace = f"{ROOT_PACKAGE_NAME}.components.{self.name}"
        return f"{namespace}.component"

    def get_provider_py_path(self, name: str) -> str:
        provider_spec = self.get_provider(name=name)
        if provider_spec.py is not None and provider_spec.py.path is not None:
            return provider_spec.py.path
        if self.kind == Kind.CUSTOM or provider_spec.kind == Kind.CUSTOM:
            return f"./providers/{provider_spec.name}.py"
        namespace = f"{ROOT_PACKAGE_NAME}.components.{self.name}"
        return f"{namespace}.providers.{provider_spec.name}"



__all__ = ["DataModel", "DataModelField"]

from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field


class DataModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    def to_dict(self):
        return self.model_dump()

    def to_json(self):
        return self.model_dump_json()

    def copy(self, deep: bool = False, **kwargs):
        return self.model_copy(deep=deep, **kwargs)

    @classmethod
    def from_dict(cls, obj: dict | None) -> Self:
        return cls.model_validate(obj)

    @classmethod
    def from_json(cls, json: str) -> Self:
        return cls.model_validate_json(json)


def DataModelField(
    alias: str | None = None,
    exclude: bool | None = None,
    **kwargs,
) -> Any:
    return Field(alias=alias, exclude=exclude, **kwargs)



import time
import uuid
from typing import Any

from ._models import BaseContext, Context, Operation, Response
from ._provider import Provider


class Component:
    provider: Provider
    _debug: bool
    _unpack: bool
    _handle: str | None

    def __init__(
        self,
        **kwargs,
    ):
        self._debug = kwargs.pop("_debug", False)
        self._unpack = kwargs.pop("_unpack", False)
        self._handle = None
        self.provider = None
        provider = kwargs.pop("provider", "default")
        self.attach(provider=provider)

    def attach(
        self,
        provider: Provider | dict | str,
    ) -> None:
        if isinstance(provider, Provider):
            self.provider = provider
        else:
            if isinstance(provider, dict):
                name = provider.pop("name")
                parameters = provider.pop("parameters", dict())
            elif isinstance(provider, str):
                name = provider
                parameters = dict()
            from ._loader import Loader

            module_name = self.__class__.__module__.rsplit(".", 1)[0]
            provider_path = f"{module_name}.providers.{name}"
            provider_instance = Loader.load_provider_instance(
                path=provider_path,
                parameters=parameters,
            )
            self.attach(provider=provider_instance)
        component_parameters = self.get_component_parameters()
        for parameter in component_parameters:
            setattr(self.provider, parameter, getattr(self, parameter))

    def init(self, context: Context | None = None) -> None:
        self.provider.init(context=context)

    async def ainit(self, context: Context | None = None) -> None:
        await self.provider.ainit(context=context)

    def run(
        self,
        operation: dict | str | Operation | None = None,
        context: dict | Context | None = None,
        **kwargs,
    ) -> Any:
        parent_context = self._convert_context(context)
        current_context = self._init_context(
            parent_context, kwargs.pop("context_info", None)
        )
        response = self.provider.run(
            operation=self._convert_operation(operation),
            context=current_context,
        )
        if isinstance(response, Response):
            if response.context is not None:
                current_context = response.context
        else:
            response = Response(result=response)
        self._finalize_context(parent_context, current_context)
        response.context = parent_context
        if not self._debug:
            response.native = None
        if self._unpack:
            return response.result
        return response

    async def arun(
        self,
        operation: dict | str | Operation | None = None,
        context: dict | Context | None = None,
        **kwargs,
    ) -> Any:
        parent_context = self._convert_context(context)
        current_context = self._init_context(
            parent_context, kwargs.pop("context_info", None)
        )
        response = await self.provider.arun(
            operation=self._convert_operation(operation),
            context=current_context,
        )
        if isinstance(response, Response):
            if response.context is not None:
                current_context = response.context
        else:
            response = Response(result=response)
        self._finalize_context(parent_context, current_context)
        response.context = parent_context
        if not self._debug:
            response.native = None
        if self._unpack:
            return response.result
        return response

    def execute(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        operation = Operation(
            name="EXECUTE",
            args=dict(statement=statement, params=params),
        )
        return self.run(operation)

    async def aexecute(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        operation = Operation(
            name="EXECUTE",
            args=dict(statement=statement, params=params),
        )
        return await self.arun(operation)

    def get_component_parameters(self):
        return [
            name
            for name in self.__annotations__.keys()
            if not name.startswith("_") and name != "provider"
        ]

    def _run_internal(
        self,
        op: str | None,
        args: dict,
        unpack_result: bool = False,
    ) -> Any:
        context = None
        unpack = False
        if "kwargs" in args:
            context = args["kwargs"].pop("context", None)
            unpack = args["kwargs"].pop("_unpack", False)
        res = self.run(
            operation=Operation.normalize(name=op, args=args), context=context
        )
        unpack = unpack or unpack_result
        if unpack:
            return res.result
        return res

    async def _arun_internal(
        self,
        op: str | None,
        args: dict,
        unpack_result: bool = False,
    ) -> Any:
        context = None
        unpack = False
        if "kwargs" in args:
            context = args["kwargs"].pop("context", None)
            unpack = args["kwargs"].pop("_unpack", False)
        res = await self.arun(
            operation=Operation.normalize(name=op, args=args), context=context
        )
        unpack = unpack or unpack_result
        if unpack:
            return res.result
        return res

    def _convert_operation(
        self,
        operation: dict | str | Operation | None,
    ) -> Operation | None:
        if isinstance(operation, dict):
            return Operation.from_dict(operation)
        elif isinstance(operation, str):
            return Operation(name=operation)
        return operation

    def _convert_context(
        self,
        context: dict | Context | None,
    ) -> Context | None:
        if isinstance(context, dict):
            return Context.from_dict(context)
        if context is None:
            return None
        return context

    def _init_context(
        self,
        parent_context: Context | None,
        context_info: dict | None,
    ) -> Context:
        provider_name = self.provider.__module__.split(".")[-1]
        component_name = self.__module__.split(".")[-2]
        if parent_context is not None:
            run_id = parent_context.run_id
        else:
            run_id = str(uuid.uuid4())
        ctx = Context(
            run_id=run_id,
            component=component_name,
            provider=provider_name,
            handle=self._handle,
            start_time=time.time(),
            parent=parent_context,
        )
        if parent_context is not None and parent_context.info is not None:
            ctx.info = parent_context.info
        if context_info is not None and ctx.info is None:
            ctx.info = context_info
        elif context_info is not None and ctx.info is not None:
            ctx.info = ctx.info.__dict__ | context_info
        return ctx

    def _finalize_context(
        self,
        parent_context: Context | None,
        current_context: Context,
    ):
        end_time = time.time()
        if current_context.start_time:
            current_context.latency = end_time - current_context.start_time
        if parent_context is not None:
            parent_context.children.append(
                BaseContext(
                    component=current_context.component,
                    provider=current_context.provider,
                    handle=current_context.handle,
                    info=current_context.info,
                    start_time=current_context.start_time,
                    latency=current_context.latency,
                    children=current_context.children,
                )
            )



from __future__ import annotations

from typing import Any

from ._models import DataModel, Operation


class OperationParser:
    operation: Operation | None

    def __init__(self, operation: Operation | None):
        self.operation = operation

    def op_equals(self, name: str | None) -> bool:
        operation_name = self.get_name()
        if operation_name is not None:
            operation_name = operation_name.lower()
        arg_name = name
        if arg_name is not None:
            arg_name = arg_name.lower()
        return operation_name == arg_name

    def get_args(self) -> dict:
        if self.operation is not None and self.operation.args is not None:
            return self.operation.args
        return {}

    def get_nargs(self) -> dict | None:
        return self.get_arg("nargs")

    def get_arg(self, arg: str, convert_dict: bool = False) -> Any:
        if (
            self.operation is not None
            and self.operation.args is not None
            and arg in self.operation.args
        ):
            arg_val = self.operation.args[arg]
            if convert_dict and isinstance(arg_val, DataModel):
                return arg_val.to_dict()
            return arg_val
        return None

    def set_arg(self, arg: str, value: Any) -> None:
        if self.operation is not None:
            if self.operation.args is None:
                self.operation.args = dict()
            self.operation.args[arg] = value

    def get_name(self) -> str | None:
        if self.operation is not None:
            return self.operation.name
        return None

    def get_statement(self) -> str | None:
        return self.get_arg("statement")



from ._component import Component
from ._decorators import operation
from ._dot_dict import DotDict
from ._loader import Loader
from ._models import BaseContext, Context, Operation, Response
from ._ncall import NCall
from ._operation_parser import OperationParser
from ._provider import Provider
from .data_model import DataModel

__all__ = [
    "BaseContext",
    "Component",
    "Context",
    "DataModel",
    "DotDict",
    "Loader",
    "NCall",
    "Operation",
    "OperationParser",
    "Provider",
    "Response",
    "operation",
]



from typing import Any

import httpx

from verse.core import Context, DataModel, Operation, Provider, Response
from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    ForbiddenError,
    InternalError,
    NotFoundError,
    NotModified,
    NotSupportedError,
    PreconditionFailedError,
    UnauthorizedError,
)


class PostRequest(DataModel):
    operation: Operation | None
    context: Context | None


class Hosted(Provider):
    endpoint: str
    api_key: str | None
    timeout: float | None
    nparams: dict[str, Any]

    def __init__(
        self,
        endpoint: str,
        api_key: str | None = None,
        timeout: float | None = 60,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            endpoint:
                HTTP endpoint.
            api_key:
                API key.
            timeout:
                HTTP timeout. Defaults to 60 seconds.
            nparams:
                Native params to httpx client.
        """
        self.endpoint = endpoint
        self.api_key = api_key
        self.timeout = timeout
        self.nparams = nparams

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Response[Any]:
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        request = PostRequest(operation=operation, context=context)
        with httpx.Client(timeout=self.timeout) as client:
            try:
                response = client.post(
                    self.endpoint,
                    json=request.to_dict(),
                    headers=headers,
                )
                response.raise_for_status()
                return Response(**response.json())
            except httpx.HTTPStatusError as e:
                raise self.handle_http_error(e.response)

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Response[Any]:
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        request = PostRequest(operation=operation, context=context)
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                response = await client.post(
                    self.endpoint,
                    json=request.to_dict(),
                    headers=headers,
                )
                response.raise_for_status()
                return Response(**response.json())
            except httpx.HTTPStatusError as e:
                raise self.handle_http_error(e.response)

    def handle_http_error(self, response: httpx.Response) -> Exception:
        if response.status_code == 304:
            return NotModified(response.text)
        if response.status_code == 400:
            return BadRequestError(response.text)
        elif response.status_code == 401:
            return UnauthorizedError(response.text)
        elif response.status_code == 403:
            return ForbiddenError(response.text)
        elif response.status_code == 404:
            return NotFoundError(response.text)
        elif response.status_code == 409:
            return ConflictError(response.text)
        elif response.status_code == 412:
            return PreconditionFailedError(response.text)
        elif response.status_code == 415:
            return NotSupportedError(response.text)
        elif response.status_code == 500:
            return InternalError(response.text)
        return InternalError(f"{response.status_code}: {response.text}")



from __future__ import annotations

from typing import Any, Callable


class NCall:
    function: Callable
    args: dict[str, Any] | list[Any] | None
    nargs: dict[str, Any] | None
    error_map: dict[Any, Any] | None

    def __init__(
        self,
        function: Callable,
        args: dict[str, Any] | list | None = None,
        nargs: dict[str, Any] | None = None,
        error_map: dict[Any, Any] | None = None,
    ):
        self.function = function
        self.args = args
        self.nargs = nargs
        self.error_map = error_map

    def __repr__(self) -> str:
        return str(self.function)

    def get_arg(self, name: str) -> Any:
        if (
            self.args is not None
            and isinstance(self.args, dict)
            and name in self.args
        ):
            return self.args[name]
        return None

    def set_arg(self, name: str, value: Any) -> None:
        if self.args is None:
            self.args = dict()
        if isinstance(self.args, dict):
            self.args[name] = value

    def invoke(self, return_error: bool = False) -> Any:
        args = self.args if self.args is not None else dict()
        nargs = self.nargs if self.nargs is not None else dict()
        if isinstance(args, dict):
            kwargs = args | nargs
        else:
            kwargs = nargs
        try:
            if not args and not kwargs:
                result = self.function()
            if isinstance(args, dict):
                result = self.function(**kwargs)
            elif isinstance(args, list):
                result = self.function(*args, **kwargs)
            if (
                result is None
                and self.error_map is not None
                and None in self.error_map
                and self.error_map[None] is not None
            ):
                raise self.error_map[None]
            return (result, None) if return_error else result
        except Exception as e:
            if self.error_map is not None:
                key = type(e)
                if key in self.error_map:
                    if self.error_map[key] is None:
                        return (None, e) if return_error else None
                    else:
                        raise self.error_map[key]
            elif return_error:
                return (None, e)
            raise

    async def ainvoke(self, return_error: bool = False) -> Any:
        args = self.args if self.args is not None else dict()
        nargs = self.nargs if self.nargs is not None else dict()
        if isinstance(args, dict):
            kwargs = args | nargs
        else:
            kwargs = nargs
        try:
            if not args and not kwargs:
                result = await self.function()
            if isinstance(args, dict):
                result = await self.function(**kwargs)
            elif isinstance(args, list):
                result = await self.function(*args, **kwargs)
            if (
                result is None
                and self.error_map is not None
                and None in self.error_map
                and self.error_map[None] is not None
            ):
                raise self.error_map[None]
            return (result, None) if return_error else result
        except Exception as e:
            if self.error_map is not None:
                key = type(e)
                if key in self.error_map:
                    if self.error_map[key] is None:
                        return (None, e) if return_error else None
                    else:
                        raise self.error_map[key]
            elif return_error:
                return (None, e)
            raise
